# -*- coding: utf-8 -*-

from __future__ import print_function

import os

import barcode

from argparse import ArgumentParser

from barcode.writer import ImageWriter, SVGWriter

# Optional PyQt4 GUI
try:
    from PyQt4 import QtCore
except ImportError:
    QtCore = None  # lint:ok

# No GUI available yet
QtCore = None
IMG_FORMATS = ('BMP', 'GIF', 'JPEG', 'MSP', 'PCX', 'PNG', 'TIFF', 'XBM')


def open_gui(args, parser=None):
    pass


def list_types(args, parser=None):
    print('\npyBarcode available barcode formats:')
    print(', '.join(barcode.PROVIDED_BARCODES))
    print('\n')
    print('Available image formats')
    print('Standard: svg')
    if ImageWriter is not None:
        print('PIL:', ', '.join(IMG_FORMATS))
    else:
        print('PIL: disabled')
    print('\n')


def create_barcode(args, parser):
    args.type = args.type.upper()
    if args.type != 'SVG' and args.type not in IMG_FORMATS:
        parser.error(
            'Unknown type {type}. Try list action for available '
            'types.'.format(type=args.type)
        )
    args.barcode = args.barcode.lower()
    if args.barcode not in barcode.PROVIDED_BARCODES:
        parser.error(
            'Unknown barcode {bc}. Try list action for available '
            'barcodes.'.format(bc=args.barcode)
        )
    if args.type != 'SVG':
        opts = dict(format=args.type)
        writer = ImageWriter()
    else:
        opts = dict(compress=args.compress)
        writer = SVGWriter()
    out = os.path.normpath(os.path.abspath(args.output))
    name = barcode.generate(args.barcode, args.code, writer, out, opts,
                            args.text)
    print('New barcode saved as {0}.'.format(name))


def main():
    msg = []
    if ImageWriter is None:
        msg.append(
            'Image output disabled (PIL not found), --type option disabled.'
        )
    else:
        msg.append(
            'Image output enabled, use --type option to give image '
            'format (png, jpeg, ...).'
        )
    if QtCore is None:
        msg.append('PyQt not found, gui action disabled.')
    else:
        msg.append('PyQt found. Use gui action to get a simple GUI.')
    parser = ArgumentParser(
        description=barcode.__description__, epilog=' '.join(msg)
    )
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + barcode.__release__)
    subparsers = parser.add_subparsers(title='Actions')
    create_parser = subparsers.add_parser('create', help='Create a barcode '
                                          'with the given options.')
    create_parser.add_argument('code', help='Code to render as barcode.')
    create_parser.add_argument('output', help='Filename for output '
                               'without extension, e. g. mybarcode.')
    create_parser.add_argument(
        '-c', '--compress', action='store_true',
        help='Compress output, only recognized if type is svg.'
    )
    create_parser.add_argument('-b', '--barcode', help='Barcode to use '
                               '[default: %(default)s].')
    create_parser.add_argument('--text', help='Text to show under the '
                               'barcode.')
    if ImageWriter is not None:
        create_parser.add_argument('-t', '--type', help='Type of output '
                                   '[default: %(default)s].')
    list_parser = subparsers.add_parser('list', help='List available '
                                        'image and code types.')
    list_parser.set_defaults(func=list_types)
    if QtCore is not None:
        gui_parser = subparsers.add_parser('gui', help='Opens a simple '
                                           'PyQt GUI to create barcodes.')
        gui_parser.set_defaults(func=open_gui)
    create_parser.set_defaults(type='svg', compress=False, func=create_barcode,
                               barcode='code39', text=None)
    args = parser.parse_args()
    args.func(args, parser)


if __name__ == '__main__':
    main()



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib.parse
from re import escape, sub
from typing import Dict, Sequence

from wrapt import ObjectProxy

from opentelemetry import context, trace

# pylint: disable=unused-import
# pylint: disable=E0611
from opentelemetry.context import _SUPPRESS_INSTRUMENTATION_KEY  # noqa: F401
from opentelemetry.propagate import extract
from opentelemetry.trace import StatusCode
from opentelemetry.trace.propagation.tracecontext import (
    TraceContextTextMapPropagator,
)

propagator = TraceContextTextMapPropagator()


def extract_attributes_from_object(
    obj: any, attributes: Sequence[str], existing: Dict[str, str] = None
) -> Dict[str, str]:
    extracted = {}
    if existing:
        extracted.update(existing)
    for attr in attributes:
        value = getattr(obj, attr, None)
        if value is not None:
            extracted[attr] = str(value)
    return extracted


def http_status_to_status_code(
    status: int,
    allow_redirect: bool = True,
    server_span: bool = False,
) -> StatusCode:
    """Converts an HTTP status code to an OpenTelemetry canonical status code

    Args:
        status (int): HTTP status code
    """
    # See: https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/http.md#status
    if not isinstance(status, int):
        return StatusCode.UNSET

    if status < 100:
        return StatusCode.ERROR
    if status <= 299:
        return StatusCode.UNSET
    if status <= 399 and allow_redirect:
        return StatusCode.UNSET
    if status <= 499 and server_span:
        return StatusCode.UNSET
    return StatusCode.ERROR


def unwrap(obj, attr: str):
    """Given a function that was wrapped by wrapt.wrap_function_wrapper, unwrap it

    Args:
        obj: Object that holds a reference to the wrapped function
        attr (str): Name of the wrapped function
    """
    func = getattr(obj, attr, None)
    if func and isinstance(func, ObjectProxy) and hasattr(func, "__wrapped__"):
        setattr(obj, attr, func.__wrapped__)


def _start_internal_or_server_span(
    tracer,
    span_name,
    start_time,
    context_carrier,
    context_getter,
    attributes=None,
):
    """Returns internal or server span along with the token which can be used by caller to reset context


    Args:
        tracer : tracer in use by given instrumentation library
        span_name (string): name of the span
        start_time : start time of the span
        context_carrier : object which contains values that are
            used to construct a Context. This object
            must be paired with an appropriate getter
            which understands how to extract a value from it.
        context_getter : an object which contains a get function that can retrieve zero
            or more values from the carrier and a keys function that can get all the keys
            from carrier.
    """

    token = ctx = span_kind = None
    if trace.get_current_span() is trace.INVALID_SPAN:
        ctx = extract(context_carrier, getter=context_getter)
        token = context.attach(ctx)
        span_kind = trace.SpanKind.SERVER
    else:
        ctx = context.get_current()
        span_kind = trace.SpanKind.INTERNAL
    span = tracer.start_span(
        name=span_name,
        context=ctx,
        kind=span_kind,
        start_time=start_time,
        attributes=attributes,
    )
    return span, token


def _url_quote(s) -> str:  # pylint: disable=invalid-name
    if not isinstance(s, (str, bytes)):
        return s
    quoted = urllib.parse.quote(s)
    # Since SQL uses '%' as a keyword, '%' is a by-product of url quoting
    # e.g. foo,bar --> foo%2Cbar
    # thus in our quoting, we need to escape it too to finally give
    #      foo,bar --> foo%%2Cbar
    return quoted.replace("%", "%%")


def _get_opentelemetry_values() -> dict:
    """
    Return the OpenTelemetry Trace and Span IDs if Span ID is set in the
    OpenTelemetry execution context.
    """
    # Insert the W3C TraceContext generated
    _headers = {}
    propagator.inject(_headers)
    return _headers


def _python_path_without_directory(python_path, directory, path_separator):
    return sub(
        rf"{escape(directory)}{path_separator}(?!$)",
        "",
        python_path,
    )



# orgnr.py - functions for handling Swedish organisation numbers
# coding: utf-8
#
# Copyright (C) 2012-2015 Arthur de Jong
# Copyright (C) 2014 Tomas Thor Jonsson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Orgnr (Organisationsnummer, Swedish company number).

The Orgnr (Organisationsnummer) is the national number to identify Swedish
companies and consists of 10 digits. These are the first 10 digits in the
Swedish VAT number, i.e. it's the VAT number without the 'SE' in front and
the '01' at the end.

>>> validate('1234567897')
'1234567897'
>>> validate('1234567891')  # invalid check digits
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('123456-7897')
'123456-7897'
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -.').strip()


def validate(number):
    """Check if the number is a valid organisation number. This checks
    the length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 10:
        raise InvalidLength()
    return luhn.validate(number)


def is_valid(number):
    """Check if the number is a valid organisation number"""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return number[:6] + '-' + number[6:]



from abc import abstractmethod
from typing import (
    Any,
    Dict,
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

from common import FiscalOrderTotals
from common.builder_functions import (
    format_datetime,
    get_optional_customer_doc,
    get_tender_description,
    pad_access_key,
    optional,
)
from nreports.builders import DataBuilder
from nreports_api.shared import JSON


class BaseFiscalNfce(DataBuilder):

    @abstractmethod
    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        raise NotImplementedError()

    @staticmethod
    def _build_header(
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
        logo_image: str = "",
    ) -> None:

        if logo_image:
            data_to_build["STORE_IMG_LOGO"] = logo_image

        data_to_build["STORE_NAME"] = fiscal_xml.find(".//emit/xNome").text
        data_to_build["STORE_DOC"] = fiscal_xml.find(".//emit/CNPJ").text
        data_to_build["STORE_STATE_INSCRIPTION"] = fiscal_xml.find(".//emit/IE").text
        data_to_build["STORE_ADDRESS"] = fiscal_xml.find(".//enderEmit/xLgr").text
        data_to_build["STORE_NUMBER"] = optional(element=fiscal_xml.find(".//enderEmit/nro")) or ""
        data_to_build["STORE_COMPLEMENT"] = optional(element=fiscal_xml.find(".//enderEmit/xCpl")) or ""
        data_to_build["STORE_NEIGHBORHOOD"] = fiscal_xml.find(".//enderEmit/xBairro").text
        data_to_build["STORE_CITY"] = fiscal_xml.find(".//enderEmit/xMun").text
        data_to_build["STORE_UF"] = fiscal_xml.find(".//enderEmit/UF").text
        data_to_build["STORE_ZIP_CODE"] = fiscal_xml.find(".//emit/enderEmit/CEP").text

    def _build_body(
        self,
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
        fiscal_order_totals: FiscalOrderTotals,
        products: List[Dict[str, str]],
    ) -> None:

        data_to_build["ORDER_ITEMS_QTY"] = len(fiscal_xml.findall(".//det"))
        data_to_build["ORDER_TOTAL_GROSS"] = fiscal_order_totals.total_value
        data_to_build["ORDER_DISCOUNT"] = fiscal_order_totals.total_discount
        data_to_build["ORDER_ADDITION"] = fiscal_order_totals.total_addition
        data_to_build["ORDER_DUE_AMOUNT"] = fiscal_xml.find(".//ICMSTot/vNF").text
        data_to_build["ORDER_CHANGE"] = optional(element=fiscal_xml.find(".//vTroco"))
        data_to_build["CUSTOMER_DOCUMENT"] = get_optional_customer_doc(fiscal_xml=fiscal_xml)
        data_to_build["ITEM_PRODUCT"] = products
        data_to_build["TENDER"] = self._get_tenders(fiscal_xml=fiscal_xml)

    def _build_footer(
        self,
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
    ) -> None:

        is_contingency_enabled = fiscal_xml.find(".//tpEmis").text != "1"
        inf_cpl = optional(element=fiscal_xml.find(".//infCpl")) or ""
        inf_cpl = "\n\n".join(inf_cpl.split("///"))
        data_to_build["NFCE_CONTINGENCY_ENABLED"] = is_contingency_enabled
        data_to_build["NFCE_ADDITIONAL_INFORMATION"] = inf_cpl
        data_to_build["NFCE_TEST_ENVIRONMENT"] = fiscal_xml.find(".//tpAmb").text == "2"
        data_to_build["NFCE_URL_CONSULT"] = fiscal_xml.find(".//urlChave").text
        data_to_build["NFCE_KEY"] = pad_access_key(access_key=fiscal_xml.find(".//infNFe").attrib["Id"][3:])
        data_to_build["NFCE_QRCODE_URL"] = fiscal_xml.find(".//qrCode").text
        data_to_build["NFCE_NUMBER"] = fiscal_xml.find(".//nNF").text
        data_to_build["NFCE_SERIES"] = fiscal_xml.find(".//serie").text
        data_to_build["NFCE_EMISSION_DATE"] = format_datetime(datetime_string=fiscal_xml.find(".//dhEmi").text)
        data_to_build["NFCE_EMISSION_OWNER"] = self._get_emission_owner(is_contingency_enabled=is_contingency_enabled)
        data_to_build["NFCE_AUTHORIZATION_PROTOCOL"] = optional(element=fiscal_xml.find(".//nProt"))
        data_to_build["NFCE_AUTHORIZATION_DATE"] = format_datetime(
            datetime_string=optional(
                element=fiscal_xml.find(".//dhRecbto"),
            ),
        )

    @staticmethod
    def _get_emission_owner(
        is_contingency_enabled: bool = False,
    ) -> Optional[List[str]]:

        if not is_contingency_enabled:
            return None

        return ["Consumidor", "Estabelecimento"]

    @staticmethod
    def _get_tenders(
        fiscal_xml: eTree.Element,
    ) -> List[Dict[str, str]]:

        tenders: List[Dict[str, str]] = []
        fiscal_tenders = fiscal_xml.findall(".//detPag")
        for tender in fiscal_tenders:
            tenders.append(
                {
                    "TENDER_TYPE": get_tender_description(tender_id=tender.find("tPag").text).upper(),
                    "TENDER_VALUE": tender.find("vPag").text,
                }
            )

        return tenders



from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates, TagEventType


def _was_served(order):
    return any(s.prod_state in (ProdStates.SERVED, ProdStates.DELIVERED) for s in order.prod_state_history)


def _is_not_served(order):
    return order.prod_state not in (ProdStates.SERVED, ProdStates.DELIVERED)


class RecalledOrderStampBox(OrderChangerProductionBox):
    def __init__(self, name, sons, filter_tag='done', ignore_prod_state=False, logger=None):
        super(RecalledOrderStampBox, self).__init__(name, sons, logger)
        self.filter_tag = filter_tag
        self.ignore_prod_state = ignore_prod_state

    def check_item_recalled(self, order, item):
        self.logger.info("Item: {}".format(item))
        if item.qty == 0 and self.filter_tag in item.tags:
            return None
        for tag_event in item.tag_history:
            if tag_event.tag in self.filter_tag and tag_event.action == TagEventType.removed:
                self.logger.info("Order marked with recalled state: {}".format(order.order_id))
                if not self.ignore_prod_state:
                    order.prod_state = ProdStates.NORMAL
                order.recalled = True
                return order

        for sub_item in item.items:
            ret = self.check_item_recalled(order, sub_item)
            if ret is not None:
                return order

        if _is_not_served(order) and _was_served(order) and not self.ignore_prod_state:
            order.recalled = True
            return order

    def change_order(self, order):
        for item in order.items:
            ret = self.check_item_recalled(order, item)
            if ret is not None:
                return order
        return order



from __future__ import annotations

from .py38 import removeprefix


def consolidate_linker_args(args: list[str]) -> str:
    """
    Ensure the return value is a string for backward compatibility.

    Retain until at least 2024-04-31. See pypa/distutils#246
    """

    if not all(arg.startswith('-Wl,') for arg in args):
        return args
    return '-Wl,' + ','.join(removeprefix(arg, '-Wl,') for arg in args)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The base interface of RPC Framework.

Implementations of this interface support the conduct of "operations":
exchanges between two distinct ends of an arbitrary number of data payloads
and metadata such as a name for the operation, initial and terminal metadata
in each direction, and flow control. These operations may be used for transfers
of data, remote procedure calls, status indication, or anything else
applications choose.
"""

# threading is referenced from specification in this module.
import abc
import enum
import threading  # pylint: disable=unused-import

# pylint: disable=too-many-arguments


class NoSuchMethodError(Exception):
    """Indicates that an unrecognized operation has been called.

    Attributes:
      code: A code value to communicate to the other side of the operation
        along with indication of operation termination. May be None.
      details: A details value to communicate to the other side of the
        operation along with indication of operation termination. May be None.
    """

    def __init__(self, code, details):
        """Constructor.

        Args:
          code: A code value to communicate to the other side of the operation
            along with indication of operation termination. May be None.
          details: A details value to communicate to the other side of the
            operation along with indication of operation termination. May be None.
        """
        super(NoSuchMethodError, self).__init__()
        self.code = code
        self.details = details


class Outcome(object):
    """The outcome of an operation.

    Attributes:
      kind: A Kind value coarsely identifying how the operation terminated.
      code: An application-specific code value or None if no such value was
        provided.
      details: An application-specific details value or None if no such value was
        provided.
    """

    @enum.unique
    class Kind(enum.Enum):
        """Ways in which an operation can terminate."""

        COMPLETED = "completed"
        CANCELLED = "cancelled"
        EXPIRED = "expired"
        LOCAL_SHUTDOWN = "local shutdown"
        REMOTE_SHUTDOWN = "remote shutdown"
        RECEPTION_FAILURE = "reception failure"
        TRANSMISSION_FAILURE = "transmission failure"
        LOCAL_FAILURE = "local failure"
        REMOTE_FAILURE = "remote failure"


class Completion(abc.ABC):
    """An aggregate of the values exchanged upon operation completion.

    Attributes:
      terminal_metadata: A terminal metadata value for the operaton.
      code: A code value for the operation.
      message: A message value for the operation.
    """


class OperationContext(abc.ABC):
    """Provides operation-related information and action."""

    @abc.abstractmethod
    def outcome(self):
        """Indicates the operation's outcome (or that the operation is ongoing).

        Returns:
          None if the operation is still active or the Outcome value for the
            operation if it has terminated.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def add_termination_callback(self, callback):
        """Adds a function to be called upon operation termination.

        Args:
          callback: A callable to be passed an Outcome value on operation
            termination.

        Returns:
          None if the operation has not yet terminated and the passed callback will
            later be called when it does terminate, or if the operation has already
            terminated an Outcome value describing the operation termination and the
            passed callback will not be called as a result of this method call.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def time_remaining(self):
        """Describes the length of allowed time remaining for the operation.

        Returns:
          A nonnegative float indicating the length of allowed time in seconds
          remaining for the operation to complete before it is considered to have
          timed out. Zero is returned if the operation has terminated.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        """Cancels the operation if the operation has not yet terminated."""
        raise NotImplementedError()

    @abc.abstractmethod
    def fail(self, exception):
        """Indicates that the operation has failed.

        Args:
          exception: An exception germane to the operation failure. May be None.
        """
        raise NotImplementedError()


class Operator(abc.ABC):
    """An interface through which to participate in an operation."""

    @abc.abstractmethod
    def advance(
        self,
        initial_metadata=None,
        payload=None,
        completion=None,
        allowance=None,
    ):
        """Progresses the operation.

        Args:
          initial_metadata: An initial metadata value. Only one may ever be
            communicated in each direction for an operation, and they must be
            communicated no later than either the first payload or the completion.
          payload: A payload value.
          completion: A Completion value. May only ever be non-None once in either
            direction, and no payloads may be passed after it has been communicated.
          allowance: A positive integer communicating the number of additional
            payloads allowed to be passed by the remote side of the operation.
        """
        raise NotImplementedError()


class ProtocolReceiver(abc.ABC):
    """A means of receiving protocol values during an operation."""

    @abc.abstractmethod
    def context(self, protocol_context):
        """Accepts the protocol context object for the operation.

        Args:
          protocol_context: The protocol context object for the operation.
        """
        raise NotImplementedError()


class Subscription(abc.ABC):
    """Describes customer code's interest in values from the other side.

    Attributes:
      kind: A Kind value describing the overall kind of this value.
      termination_callback: A callable to be passed the Outcome associated with
        the operation after it has terminated. Must be non-None if kind is
        Kind.TERMINATION_ONLY. Must be None otherwise.
      allowance: A callable behavior that accepts positive integers representing
        the number of additional payloads allowed to be passed to the other side
        of the operation. Must be None if kind is Kind.FULL. Must not be None
        otherwise.
      operator: An Operator to be passed values from the other side of the
        operation. Must be non-None if kind is Kind.FULL. Must be None otherwise.
      protocol_receiver: A ProtocolReceiver to be passed protocol objects as they
        become available during the operation. Must be non-None if kind is
        Kind.FULL.
    """

    @enum.unique
    class Kind(enum.Enum):
        NONE = "none"
        TERMINATION_ONLY = "termination only"
        FULL = "full"


class Servicer(abc.ABC):
    """Interface for service implementations."""

    @abc.abstractmethod
    def service(self, group, method, context, output_operator):
        """Services an operation.

        Args:
          group: The group identifier of the operation to be serviced.
          method: The method identifier of the operation to be serviced.
          context: An OperationContext object affording contextual information and
            actions.
          output_operator: An Operator that will accept output values of the
            operation.

        Returns:
          A Subscription via which this object may or may not accept more values of
            the operation.

        Raises:
          NoSuchMethodError: If this Servicer does not handle operations with the
            given group and method.
          abandonment.Abandoned: If the operation has been aborted and there no
            longer is any reason to service the operation.
        """
        raise NotImplementedError()


class End(abc.ABC):
    """Common type for entry-point objects on both sides of an operation."""

    @abc.abstractmethod
    def start(self):
        """Starts this object's service of operations."""
        raise NotImplementedError()

    @abc.abstractmethod
    def stop(self, grace):
        """Stops this object's service of operations.

        This object will refuse service of new operations as soon as this method is
        called but operations under way at the time of the call may be given a
        grace period during which they are allowed to finish.

        Args:
          grace: A duration of time in seconds to allow ongoing operations to
            terminate before being forcefully terminated by the stopping of this
            End. May be zero to terminate all ongoing operations and immediately
            stop.

        Returns:
          A threading.Event that will be set to indicate all operations having
            terminated and this End having completely stopped. The returned event
            may not be set until after the full grace period (if some ongoing
            operation continues for the full length of the period) or it may be set
            much sooner (if for example this End had no operations in progress at
            the time its stop method was called).
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def operate(
        self,
        group,
        method,
        subscription,
        timeout,
        initial_metadata=None,
        payload=None,
        completion=None,
        protocol_options=None,
    ):
        """Commences an operation.

        Args:
          group: The group identifier of the invoked operation.
          method: The method identifier of the invoked operation.
          subscription: A Subscription to which the results of the operation will be
            passed.
          timeout: A length of time in seconds to allow for the operation.
          initial_metadata: An initial metadata value to be sent to the other side
            of the operation. May be None if the initial metadata will be later
            passed via the returned operator or if there will be no initial metadata
            passed at all.
          payload: An initial payload for the operation.
          completion: A Completion value indicating the end of transmission to the
            other side of the operation.
          protocol_options: A value specified by the provider of a Base interface
            implementation affording custom state and behavior.

        Returns:
          A pair of objects affording information about the operation and action
            continuing the operation. The first element of the returned pair is an
            OperationContext for the operation and the second element of the
            returned pair is an Operator to which operation values not passed in
            this call should later be passed.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def operation_stats(self):
        """Reports the number of terminated operations broken down by outcome.

        Returns:
          A dictionary from Outcome.Kind value to an integer identifying the number
            of operations that terminated with that outcome kind.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def add_idle_action(self, action):
        """Adds an action to be called when this End has no ongoing operations.

        Args:
          action: A callable that accepts no arguments.
        """
        raise NotImplementedError()



import json
import logging

from bustoken import TK_LOYALTY_GET_ENDPOINTS_CONFIG
from msgbus import (
    FM_PARAM,
    MBException,
    SE_NOTFOUND,
    TK_SYS_ACK,
)
from sysactions import send_message
from actions.models import LoyaltyEndpointsConfig

logger = logging.getLogger("PosActions")
loyalty_endpoints_config = None


def should_ask_loyalty_message_on_total():
    # type: (...) -> bool

    return _get_loyalty_message_config(ask_on="ask_totaled")


def should_ask_loyalty_on_payment():
    # type: (...) -> bool

    return _get_loyalty_message_config(ask_on="ask_payment")


def _get_loyalty_message_config(ask_on):
    # type: (...) -> bool

    global loyalty_endpoints_config

    if loyalty_endpoints_config is not None:
        return getattr(loyalty_endpoints_config, ask_on)

    try:
        logger.info("Start getting loyalty endpoints config")
        response = send_message(
            dest_name="Loyalty",
            token=TK_LOYALTY_GET_ENDPOINTS_CONFIG,
            format=FM_PARAM,
        )

        if response.token == TK_SYS_ACK:
            response_dict = json.loads(response.data)
            loyalty_endpoints_config = LoyaltyEndpointsConfig(
                ask_totaled=response_dict.get("AskTotaled"),
                ask_payment=response_dict.get("AskPayment"),
            )
        else:
            logger.info("TK_LOYALTY_GET_ENDPOINTS_CONFIG returned NAK")

        return getattr(loyalty_endpoints_config, ask_on)
    except MBException as ex:
        if ex.errorcode == SE_NOTFOUND:
            logger.info("Component Loyalty not alive")
            loyalty_endpoints_config = LoyaltyEndpointsConfig(
                ask_totaled=False,
                ask_payment=False,
            )
    except (Exception,):
        logger.exception("Error getting loyalty endpoint config")

    return False



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from opentelemetry.sdk.metrics._internal.export import (
    AggregationTemporality,
    ConsoleMetricExporter,
    InMemoryMetricReader,
    MetricExporter,
    MetricExportResult,
    MetricReader,
    PeriodicExportingMetricReader,
)

# The point module is not in the export directory to avoid a circular import.
from opentelemetry.sdk.metrics._internal.point import (  # noqa: F401
    Buckets,
    DataPointT,
    DataT,
    ExponentialHistogram,
    ExponentialHistogramDataPoint,
    Gauge,
    Histogram,
    HistogramDataPoint,
    Metric,
    MetricsData,
    NumberDataPoint,
    ResourceMetrics,
    ScopeMetrics,
    Sum,
)

__all__ = [
    "AggregationTemporality",
    "ConsoleMetricExporter",
    "InMemoryMetricReader",
    "MetricExporter",
    "MetricExportResult",
    "MetricReader",
    "PeriodicExportingMetricReader",
    "DataPointT",
    "DataT",
    "Gauge",
    "Histogram",
    "HistogramDataPoint",
    "Metric",
    "MetricsData",
    "NumberDataPoint",
    "ResourceMetrics",
    "ScopeMetrics",
    "Sum",
]



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opentelemetry import context
from opentelemetry.instrumentation.utils import _url_quote


def _add_sql_comment(sql, **meta) -> str:
    """
    Appends comments to the sql statement and returns it
    """
    meta.update(**_add_framework_tags())
    comment = _generate_sql_comment(**meta)
    sql = sql.rstrip()
    if sql[-1] == ";":
        sql = sql[:-1] + comment + ";"
    else:
        sql = sql + comment
    return sql


def _generate_sql_comment(**meta) -> str:
    """
    Return a SQL comment with comma delimited key=value pairs created from
    **meta kwargs.
    """
    key_value_delimiter = ","

    if not meta:  # No entries added.
        return ""

    # Sort the keywords to ensure that caching works and that testing is
    # deterministic. It eases visual inspection as well.
    return (
        " /*"
        + key_value_delimiter.join(
            f"{_url_quote(key)}={_url_quote(value)!r}"
            for key, value in sorted(meta.items())
            if value is not None
        )
        + "*/"
    )


def _add_framework_tags() -> dict:
    """
    Returns orm related tags if any set by the context
    """

    sqlcommenter_framework_values = (
        context.get_value("SQLCOMMENTER_ORM_TAGS_AND_VALUES")
        if context.get_value("SQLCOMMENTER_ORM_TAGS_AND_VALUES")
        else {}
    )
    return sqlcommenter_framework_values



# personnummer.py - functions for handling Swedish Personal identity numbers
# coding: utf-8
#
# Copyright (C) 2018 Ilya Vihtinsky
# Copyright (C) 2018-2020 Arthur de Jong
# Copyright (C) 2020 Leon Sandøy
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Personnummer (Swedish personal identity number).

The Swedish Personnummer is assigned at birth to all Swedish nationals and to
immigrants for tax and identification purposes. The number consists of 10 or
12 digits and starts with the birth date, followed by a serial and a check
digit.

More information:

* https://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)

>>> validate('880320-0016')
'880320-0016'
>>> validate('880320-0018')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> get_gender('890102-3286')
'F'
>>> get_birth_date('811228-9841')
datetime.date(1981, 12, 28)
>>> format('8803200016')
'880320-0016'
"""

import datetime

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' :')
    if len(number) in (10, 12) and number[-5] not in '-+':
        number = '%s-%s' % (number[:-4], number[-4:])
    return number[:-5].replace('-', '').replace('+', '') + number[-5:]


def get_birth_date(number):
    """Determine the birth date from the number.

    For people aged 100 and up, the minus/dash in the personnummer is changed to a plus
    on New Year's Eve the year they turn 100.

    See Folkbokföringslagen (1991:481), §18.
    """
    number = compact(number)
    if len(number) == 13:
        year = int(number[0:4])
        month = int(number[4:6])
        day = int(number[6:8])
    else:
        year = datetime.date.today().year
        century = year // 100
        if int(number[0:2]) > year % 100:
            century -= 1
        if number[-5] == '+':
            century -= 1
        year = int('%d%s' % (century, number[0:2]))
        month = int(number[2:4])
        day = int(number[4:6])
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    if int(number[-2]) % 2:
        return 'M'
    else:
        return 'F'


def validate(number):
    """Check if the number is a valid identity number."""
    number = compact(number)
    if len(number) not in (11, 13):
        raise InvalidLength()
    digits = number[:-5] + number[-4:]
    if number[-5] not in '-+' or not isdigits(digits):
        raise InvalidFormat()
    get_birth_date(number)
    luhn.validate(digits[-10:])
    return number


def is_valid(number):
    """Check if the number is a valid identity number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



data = (
'geul',    # 0x00
'geulg',    # 0x01
'geulm',    # 0x02
'geulb',    # 0x03
'geuls',    # 0x04
'geult',    # 0x05
'geulp',    # 0x06
'geulh',    # 0x07
'geum',    # 0x08
'geub',    # 0x09
'geubs',    # 0x0a
'geus',    # 0x0b
'geuss',    # 0x0c
'geung',    # 0x0d
'geuj',    # 0x0e
'geuc',    # 0x0f
'geuk',    # 0x10
'geut',    # 0x11
'geup',    # 0x12
'geuh',    # 0x13
'gyi',    # 0x14
'gyig',    # 0x15
'gyigg',    # 0x16
'gyigs',    # 0x17
'gyin',    # 0x18
'gyinj',    # 0x19
'gyinh',    # 0x1a
'gyid',    # 0x1b
'gyil',    # 0x1c
'gyilg',    # 0x1d
'gyilm',    # 0x1e
'gyilb',    # 0x1f
'gyils',    # 0x20
'gyilt',    # 0x21
'gyilp',    # 0x22
'gyilh',    # 0x23
'gyim',    # 0x24
'gyib',    # 0x25
'gyibs',    # 0x26
'gyis',    # 0x27
'gyiss',    # 0x28
'gying',    # 0x29
'gyij',    # 0x2a
'gyic',    # 0x2b
'gyik',    # 0x2c
'gyit',    # 0x2d
'gyip',    # 0x2e
'gyih',    # 0x2f
'gi',    # 0x30
'gig',    # 0x31
'gigg',    # 0x32
'gigs',    # 0x33
'gin',    # 0x34
'ginj',    # 0x35
'ginh',    # 0x36
'gid',    # 0x37
'gil',    # 0x38
'gilg',    # 0x39
'gilm',    # 0x3a
'gilb',    # 0x3b
'gils',    # 0x3c
'gilt',    # 0x3d
'gilp',    # 0x3e
'gilh',    # 0x3f
'gim',    # 0x40
'gib',    # 0x41
'gibs',    # 0x42
'gis',    # 0x43
'giss',    # 0x44
'ging',    # 0x45
'gij',    # 0x46
'gic',    # 0x47
'gik',    # 0x48
'git',    # 0x49
'gip',    # 0x4a
'gih',    # 0x4b
'gga',    # 0x4c
'ggag',    # 0x4d
'ggagg',    # 0x4e
'ggags',    # 0x4f
'ggan',    # 0x50
'gganj',    # 0x51
'gganh',    # 0x52
'ggad',    # 0x53
'ggal',    # 0x54
'ggalg',    # 0x55
'ggalm',    # 0x56
'ggalb',    # 0x57
'ggals',    # 0x58
'ggalt',    # 0x59
'ggalp',    # 0x5a
'ggalh',    # 0x5b
'ggam',    # 0x5c
'ggab',    # 0x5d
'ggabs',    # 0x5e
'ggas',    # 0x5f
'ggass',    # 0x60
'ggang',    # 0x61
'ggaj',    # 0x62
'ggac',    # 0x63
'ggak',    # 0x64
'ggat',    # 0x65
'ggap',    # 0x66
'ggah',    # 0x67
'ggae',    # 0x68
'ggaeg',    # 0x69
'ggaegg',    # 0x6a
'ggaegs',    # 0x6b
'ggaen',    # 0x6c
'ggaenj',    # 0x6d
'ggaenh',    # 0x6e
'ggaed',    # 0x6f
'ggael',    # 0x70
'ggaelg',    # 0x71
'ggaelm',    # 0x72
'ggaelb',    # 0x73
'ggaels',    # 0x74
'ggaelt',    # 0x75
'ggaelp',    # 0x76
'ggaelh',    # 0x77
'ggaem',    # 0x78
'ggaeb',    # 0x79
'ggaebs',    # 0x7a
'ggaes',    # 0x7b
'ggaess',    # 0x7c
'ggaeng',    # 0x7d
'ggaej',    # 0x7e
'ggaec',    # 0x7f
'ggaek',    # 0x80
'ggaet',    # 0x81
'ggaep',    # 0x82
'ggaeh',    # 0x83
'ggya',    # 0x84
'ggyag',    # 0x85
'ggyagg',    # 0x86
'ggyags',    # 0x87
'ggyan',    # 0x88
'ggyanj',    # 0x89
'ggyanh',    # 0x8a
'ggyad',    # 0x8b
'ggyal',    # 0x8c
'ggyalg',    # 0x8d
'ggyalm',    # 0x8e
'ggyalb',    # 0x8f
'ggyals',    # 0x90
'ggyalt',    # 0x91
'ggyalp',    # 0x92
'ggyalh',    # 0x93
'ggyam',    # 0x94
'ggyab',    # 0x95
'ggyabs',    # 0x96
'ggyas',    # 0x97
'ggyass',    # 0x98
'ggyang',    # 0x99
'ggyaj',    # 0x9a
'ggyac',    # 0x9b
'ggyak',    # 0x9c
'ggyat',    # 0x9d
'ggyap',    # 0x9e
'ggyah',    # 0x9f
'ggyae',    # 0xa0
'ggyaeg',    # 0xa1
'ggyaegg',    # 0xa2
'ggyaegs',    # 0xa3
'ggyaen',    # 0xa4
'ggyaenj',    # 0xa5
'ggyaenh',    # 0xa6
'ggyaed',    # 0xa7
'ggyael',    # 0xa8
'ggyaelg',    # 0xa9
'ggyaelm',    # 0xaa
'ggyaelb',    # 0xab
'ggyaels',    # 0xac
'ggyaelt',    # 0xad
'ggyaelp',    # 0xae
'ggyaelh',    # 0xaf
'ggyaem',    # 0xb0
'ggyaeb',    # 0xb1
'ggyaebs',    # 0xb2
'ggyaes',    # 0xb3
'ggyaess',    # 0xb4
'ggyaeng',    # 0xb5
'ggyaej',    # 0xb6
'ggyaec',    # 0xb7
'ggyaek',    # 0xb8
'ggyaet',    # 0xb9
'ggyaep',    # 0xba
'ggyaeh',    # 0xbb
'ggeo',    # 0xbc
'ggeog',    # 0xbd
'ggeogg',    # 0xbe
'ggeogs',    # 0xbf
'ggeon',    # 0xc0
'ggeonj',    # 0xc1
'ggeonh',    # 0xc2
'ggeod',    # 0xc3
'ggeol',    # 0xc4
'ggeolg',    # 0xc5
'ggeolm',    # 0xc6
'ggeolb',    # 0xc7
'ggeols',    # 0xc8
'ggeolt',    # 0xc9
'ggeolp',    # 0xca
'ggeolh',    # 0xcb
'ggeom',    # 0xcc
'ggeob',    # 0xcd
'ggeobs',    # 0xce
'ggeos',    # 0xcf
'ggeoss',    # 0xd0
'ggeong',    # 0xd1
'ggeoj',    # 0xd2
'ggeoc',    # 0xd3
'ggeok',    # 0xd4
'ggeot',    # 0xd5
'ggeop',    # 0xd6
'ggeoh',    # 0xd7
'gge',    # 0xd8
'ggeg',    # 0xd9
'ggegg',    # 0xda
'ggegs',    # 0xdb
'ggen',    # 0xdc
'ggenj',    # 0xdd
'ggenh',    # 0xde
'gged',    # 0xdf
'ggel',    # 0xe0
'ggelg',    # 0xe1
'ggelm',    # 0xe2
'ggelb',    # 0xe3
'ggels',    # 0xe4
'ggelt',    # 0xe5
'ggelp',    # 0xe6
'ggelh',    # 0xe7
'ggem',    # 0xe8
'ggeb',    # 0xe9
'ggebs',    # 0xea
'gges',    # 0xeb
'ggess',    # 0xec
'ggeng',    # 0xed
'ggej',    # 0xee
'ggec',    # 0xef
'ggek',    # 0xf0
'gget',    # 0xf1
'ggep',    # 0xf2
'ggeh',    # 0xf3
'ggyeo',    # 0xf4
'ggyeog',    # 0xf5
'ggyeogg',    # 0xf6
'ggyeogs',    # 0xf7
'ggyeon',    # 0xf8
'ggyeonj',    # 0xf9
'ggyeonh',    # 0xfa
'ggyeod',    # 0xfb
'ggyeol',    # 0xfc
'ggyeolg',    # 0xfd
'ggyeolm',    # 0xfe
'ggyeolb',    # 0xff
)



import json

from messagebus import (
    Event,
    MessageBus,
)
from typing import Any
from ._DefaultMessageProcessor import DefaultMessageProcessor
from ..interactor import RemoteLoaderUpdaterInteractor
from ..model import DispatchedEvents


class RemoteLoaderUpdaterProcessor(DefaultMessageProcessor):
    def __init__(
        self,           # type: RemoteLoaderUpdaterProcessor
        interactor,     # type: RemoteLoaderUpdaterInteractor
    ):
        # type: (...) -> None

        super(DefaultMessageProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,           # type: RemoteLoaderUpdaterProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> str

        json_data = json.loads(data)
        return json_data.get("externalUpdateId")

    def call_business(
        self,                   # type: RemoteLoaderUpdaterProcessor
        external_update_id,     # type: str
    ):
        # type: (...) -> Any

        return self.interactor.update_loaders(external_update_id=external_update_id)

    def format_response(
        self,           # type: RemoteLoaderUpdaterProcessor
        message_bus,    # type: MessageBus
        _,              # type: Any
        event,          # type: Event
        input_obj,      # type: Any
        __,             # type: Any
    ):
        # type: (...) -> None

        data = {
            "externalUpdateId": input_obj
        }
        reply_event = Event(
            subject=DispatchedEvents.EVT_POS_UPDATER_REMOTE_LOADER_UPDATE_ACK.value,
            data=json.dumps(data),
        )
        if event:
            message_bus.publish_event(event=reply_event)

    def format_exception(
        self,           # type: RemoteLoaderUpdaterProcessor
        message_bus,    # type: MessageBus
        _,              # type: Any
        event,          # type: Event
        input_obj,      # type: Any
        __,             # type: Any
    ):
        # type: (...) -> None

        return



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package implements `OpenTelemetry Resources
<https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/resource/sdk.md#resource-sdk>`_:

    *A Resource is an immutable representation of the entity producing
    telemetry. For example, a process producing telemetry that is running in
    a container on Kubernetes has a Pod name, it is in a namespace and
    possibly is part of a Deployment which also has a name. All three of
    these attributes can be included in the Resource.*

Resource objects are created with `Resource.create`, which accepts attributes
(key-values). Resources should NOT be created via constructor, and working with
`Resource` objects should only be done via the Resource API methods. Resource
attributes can also be passed at process invocation in the
:envvar:`OTEL_RESOURCE_ATTRIBUTES` environment variable. You should register
your resource with the  `opentelemetry.sdk.trace.TracerProvider` by passing
them into their constructors. The `Resource` passed to a provider is available
to the exporter, which can send on this information as it sees fit.

.. code-block:: python

    trace.set_tracer_provider(
        TracerProvider(
            resource=Resource.create({
                "service.name": "shoppingcart",
                "service.instance.id": "instance-12",
            }),
        ),
    )
    print(trace.get_tracer_provider().resource.attributes)

    {'telemetry.sdk.language': 'python',
    'telemetry.sdk.name': 'opentelemetry',
    'telemetry.sdk.version': '0.13.dev0',
    'service.name': 'shoppingcart',
    'service.instance.id': 'instance-12'}

Note that the OpenTelemetry project documents certain `"standard attributes"
<https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/resource/semantic_conventions/README.md>`_
that have prescribed semantic meanings, for example ``service.name`` in the
above example.
 """

import abc
import concurrent.futures
import logging
import sys
import typing
from json import dumps
from os import environ
from urllib import parse

from opentelemetry.attributes import BoundedAttributes
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPERIMENTAL_RESOURCE_DETECTORS,
    OTEL_RESOURCE_ATTRIBUTES,
    OTEL_SERVICE_NAME,
)
from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.util._importlib_metadata import entry_points, version
from opentelemetry.util.types import AttributeValue

LabelValue = AttributeValue
Attributes = typing.Dict[str, LabelValue]
logger = logging.getLogger(__name__)


CLOUD_PROVIDER = ResourceAttributes.CLOUD_PROVIDER
CLOUD_ACCOUNT_ID = ResourceAttributes.CLOUD_ACCOUNT_ID
CLOUD_REGION = ResourceAttributes.CLOUD_REGION
CLOUD_AVAILABILITY_ZONE = ResourceAttributes.CLOUD_AVAILABILITY_ZONE
CONTAINER_NAME = ResourceAttributes.CONTAINER_NAME
CONTAINER_ID = ResourceAttributes.CONTAINER_ID
CONTAINER_IMAGE_NAME = ResourceAttributes.CONTAINER_IMAGE_NAME
CONTAINER_IMAGE_TAG = ResourceAttributes.CONTAINER_IMAGE_TAG
DEPLOYMENT_ENVIRONMENT = ResourceAttributes.DEPLOYMENT_ENVIRONMENT
FAAS_NAME = ResourceAttributes.FAAS_NAME
FAAS_ID = ResourceAttributes.FAAS_ID
FAAS_VERSION = ResourceAttributes.FAAS_VERSION
FAAS_INSTANCE = ResourceAttributes.FAAS_INSTANCE
HOST_NAME = ResourceAttributes.HOST_NAME
HOST_TYPE = ResourceAttributes.HOST_TYPE
HOST_IMAGE_NAME = ResourceAttributes.HOST_IMAGE_NAME
HOST_IMAGE_ID = ResourceAttributes.HOST_IMAGE_ID
HOST_IMAGE_VERSION = ResourceAttributes.HOST_IMAGE_VERSION
KUBERNETES_CLUSTER_NAME = ResourceAttributes.K8S_CLUSTER_NAME
KUBERNETES_NAMESPACE_NAME = ResourceAttributes.K8S_NAMESPACE_NAME
KUBERNETES_POD_UID = ResourceAttributes.K8S_POD_UID
KUBERNETES_POD_NAME = ResourceAttributes.K8S_POD_NAME
KUBERNETES_CONTAINER_NAME = ResourceAttributes.K8S_CONTAINER_NAME
KUBERNETES_REPLICA_SET_UID = ResourceAttributes.K8S_REPLICASET_UID
KUBERNETES_REPLICA_SET_NAME = ResourceAttributes.K8S_REPLICASET_NAME
KUBERNETES_DEPLOYMENT_UID = ResourceAttributes.K8S_DEPLOYMENT_UID
KUBERNETES_DEPLOYMENT_NAME = ResourceAttributes.K8S_DEPLOYMENT_NAME
KUBERNETES_STATEFUL_SET_UID = ResourceAttributes.K8S_STATEFULSET_UID
KUBERNETES_STATEFUL_SET_NAME = ResourceAttributes.K8S_STATEFULSET_NAME
KUBERNETES_DAEMON_SET_UID = ResourceAttributes.K8S_DAEMONSET_UID
KUBERNETES_DAEMON_SET_NAME = ResourceAttributes.K8S_DAEMONSET_NAME
KUBERNETES_JOB_UID = ResourceAttributes.K8S_JOB_UID
KUBERNETES_JOB_NAME = ResourceAttributes.K8S_JOB_NAME
KUBERNETES_CRON_JOB_UID = ResourceAttributes.K8S_CRONJOB_UID
KUBERNETES_CRON_JOB_NAME = ResourceAttributes.K8S_CRONJOB_NAME
OS_TYPE = ResourceAttributes.OS_TYPE
OS_DESCRIPTION = ResourceAttributes.OS_DESCRIPTION
PROCESS_PID = ResourceAttributes.PROCESS_PID
PROCESS_EXECUTABLE_NAME = ResourceAttributes.PROCESS_EXECUTABLE_NAME
PROCESS_EXECUTABLE_PATH = ResourceAttributes.PROCESS_EXECUTABLE_PATH
PROCESS_COMMAND = ResourceAttributes.PROCESS_COMMAND
PROCESS_COMMAND_LINE = ResourceAttributes.PROCESS_COMMAND_LINE
PROCESS_COMMAND_ARGS = ResourceAttributes.PROCESS_COMMAND_ARGS
PROCESS_OWNER = ResourceAttributes.PROCESS_OWNER
PROCESS_RUNTIME_NAME = ResourceAttributes.PROCESS_RUNTIME_NAME
PROCESS_RUNTIME_VERSION = ResourceAttributes.PROCESS_RUNTIME_VERSION
PROCESS_RUNTIME_DESCRIPTION = ResourceAttributes.PROCESS_RUNTIME_DESCRIPTION
SERVICE_NAME = ResourceAttributes.SERVICE_NAME
SERVICE_NAMESPACE = ResourceAttributes.SERVICE_NAMESPACE
SERVICE_INSTANCE_ID = ResourceAttributes.SERVICE_INSTANCE_ID
SERVICE_VERSION = ResourceAttributes.SERVICE_VERSION
TELEMETRY_SDK_NAME = ResourceAttributes.TELEMETRY_SDK_NAME
TELEMETRY_SDK_VERSION = ResourceAttributes.TELEMETRY_SDK_VERSION
TELEMETRY_AUTO_VERSION = ResourceAttributes.TELEMETRY_AUTO_VERSION
TELEMETRY_SDK_LANGUAGE = ResourceAttributes.TELEMETRY_SDK_LANGUAGE


_OPENTELEMETRY_SDK_VERSION = version("opentelemetry-sdk")


class Resource:
    """A Resource is an immutable representation of the entity producing telemetry as Attributes."""

    def __init__(
        self, attributes: Attributes, schema_url: typing.Optional[str] = None
    ):
        self._attributes = BoundedAttributes(attributes=attributes)
        if schema_url is None:
            schema_url = ""
        self._schema_url = schema_url

    @staticmethod
    def create(
        attributes: typing.Optional[Attributes] = None,
        schema_url: typing.Optional[str] = None,
    ) -> "Resource":
        """Creates a new `Resource` from attributes.

        Args:
            attributes: Optional zero or more key-value pairs.
            schema_url: Optional URL pointing to the schema

        Returns:
            The newly-created Resource.
        """

        if not attributes:
            attributes = {}

        resource_detectors = []

        resource = _DEFAULT_RESOURCE

        otel_experimental_resource_detectors = environ.get(
            OTEL_EXPERIMENTAL_RESOURCE_DETECTORS, "otel"
        ).split(",")

        if "otel" not in otel_experimental_resource_detectors:
            otel_experimental_resource_detectors.append("otel")

        for resource_detector in otel_experimental_resource_detectors:

            resource_detectors.append(
                next(
                    iter(
                        entry_points(
                            group="opentelemetry_resource_detector",
                            name=resource_detector.strip(),
                        )
                    )
                ).load()()
            )

        resource = get_aggregated_resources(
            resource_detectors, _DEFAULT_RESOURCE
        ).merge(Resource(attributes, schema_url))

        if not resource.attributes.get(SERVICE_NAME, None):
            default_service_name = "unknown_service"
            process_executable_name = resource.attributes.get(
                PROCESS_EXECUTABLE_NAME, None
            )
            if process_executable_name:
                default_service_name += ":" + process_executable_name
            resource = resource.merge(
                Resource({SERVICE_NAME: default_service_name}, schema_url)
            )
        return resource

    @staticmethod
    def get_empty() -> "Resource":
        return _EMPTY_RESOURCE

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def schema_url(self) -> str:
        return self._schema_url

    def merge(self, other: "Resource") -> "Resource":
        """Merges this resource and an updating resource into a new `Resource`.

        If a key exists on both the old and updating resource, the value of the
        updating resource will override the old resource value.

        The updating resource's `schema_url` will be used only if the old
        `schema_url` is empty. Attempting to merge two resources with
        different, non-empty values for `schema_url` will result in an error
        and return the old resource.

        Args:
            other: The other resource to be merged.

        Returns:
            The newly-created Resource.
        """
        merged_attributes = self.attributes.copy()
        merged_attributes.update(other.attributes)

        if self.schema_url == "":
            schema_url = other.schema_url
        elif other.schema_url == "":
            schema_url = self.schema_url
        elif self.schema_url == other.schema_url:
            schema_url = other.schema_url
        else:
            logger.error(
                "Failed to merge resources: The two schemas %s and %s are incompatible",
                self.schema_url,
                other.schema_url,
            )
            return self

        return Resource(merged_attributes, schema_url)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Resource):
            return False
        return (
            self._attributes == other._attributes
            and self._schema_url == other._schema_url
        )

    def __hash__(self):
        return hash(
            f"{dumps(self._attributes.copy(), sort_keys=True)}|{self._schema_url}"
        )

    def to_json(self, indent=4) -> str:
        return dumps(
            {
                "attributes": dict(self._attributes),
                "schema_url": self._schema_url,
            },
            indent=indent,
        )


_EMPTY_RESOURCE = Resource({})
_DEFAULT_RESOURCE = Resource(
    {
        TELEMETRY_SDK_LANGUAGE: "python",
        TELEMETRY_SDK_NAME: "opentelemetry",
        TELEMETRY_SDK_VERSION: _OPENTELEMETRY_SDK_VERSION,
    }
)


class ResourceDetector(abc.ABC):
    def __init__(self, raise_on_error=False):
        self.raise_on_error = raise_on_error

    @abc.abstractmethod
    def detect(self) -> "Resource":
        raise NotImplementedError()


class OTELResourceDetector(ResourceDetector):
    # pylint: disable=no-self-use
    def detect(self) -> "Resource":

        env_resources_items = environ.get(OTEL_RESOURCE_ATTRIBUTES)
        env_resource_map = {}

        if env_resources_items:
            for item in env_resources_items.split(","):
                try:
                    key, value = item.split("=", maxsplit=1)
                except ValueError as exc:
                    logger.warning(
                        "Invalid key value resource attribute pair %s: %s",
                        item,
                        exc,
                    )
                    continue
                value_url_decoded = parse.unquote(value.strip())
                env_resource_map[key.strip()] = value_url_decoded

        service_name = environ.get(OTEL_SERVICE_NAME)
        if service_name:
            env_resource_map[SERVICE_NAME] = service_name
        return Resource(env_resource_map)


class ProcessResourceDetector(ResourceDetector):
    # pylint: disable=no-self-use
    def detect(self) -> "Resource":
        _runtime_version = ".".join(
            map(
                str,
                sys.version_info[:3]
                if sys.version_info.releaselevel == "final"
                and not sys.version_info.serial
                else sys.version_info,
            )
        )

        return Resource(
            {
                PROCESS_RUNTIME_DESCRIPTION: sys.version,
                PROCESS_RUNTIME_NAME: sys.implementation.name,
                PROCESS_RUNTIME_VERSION: _runtime_version,
            }
        )


def get_aggregated_resources(
    detectors: typing.List["ResourceDetector"],
    initial_resource: typing.Optional[Resource] = None,
    timeout=5,
) -> "Resource":
    """Retrieves resources from detectors in the order that they were passed

    :param detectors: List of resources in order of priority
    :param initial_resource: Static resource. This has highest priority
    :param timeout: Number of seconds to wait for each detector to return
    :return:
    """
    detectors_merged_resource = initial_resource or Resource.create()

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        futures = [executor.submit(detector.detect) for detector in detectors]
        for detector_ind, future in enumerate(futures):
            detector = detectors[detector_ind]
            try:
                detected_resource = future.result(timeout=timeout)
            # pylint: disable=broad-except
            except Exception as ex:
                detected_resource = _EMPTY_RESOURCE
                if detector.raise_on_error:
                    raise ex
                logger.warning(
                    "Exception %s in detector %s, ignoring", ex, detector
                )
            finally:
                detectors_merged_resource = detectors_merged_resource.merge(
                    detected_resource
                )

    return detectors_merged_resource



# sql/_selectable_constructors.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from . import coercions
from . import roles
from ._typing import _ColumnsClauseArgument
from ._typing import _no_kw
from .elements import ColumnClause
from .selectable import Alias
from .selectable import CompoundSelect
from .selectable import Exists
from .selectable import FromClause
from .selectable import Join
from .selectable import Lateral
from .selectable import LateralFromClause
from .selectable import NamedFromClause
from .selectable import Select
from .selectable import TableClause
from .selectable import TableSample
from .selectable import Values

if TYPE_CHECKING:
    from ._typing import _FromClauseArgument
    from ._typing import _OnClauseArgument
    from ._typing import _SelectStatementForCompoundArgument
    from ._typing import _T0
    from ._typing import _T1
    from ._typing import _T2
    from ._typing import _T3
    from ._typing import _T4
    from ._typing import _T5
    from ._typing import _T6
    from ._typing import _T7
    from ._typing import _T8
    from ._typing import _T9
    from ._typing import _TypedColumnClauseArgument as _TCCA
    from .functions import Function
    from .selectable import CTE
    from .selectable import HasCTE
    from .selectable import ScalarSelect
    from .selectable import SelectBase


_T = TypeVar("_T", bound=Any)


def alias(
    selectable: FromClause, name: Optional[str] = None, flat: bool = False
) -> NamedFromClause:
    """Return a named alias of the given :class:`.FromClause`.

    For :class:`.Table` and :class:`.Join` objects, the return type is the
    :class:`_expression.Alias` object. Other kinds of :class:`.NamedFromClause`
    objects may be returned for other kinds of :class:`.FromClause` objects.

    The named alias represents any :class:`_expression.FromClause` with an
    alternate name assigned within SQL, typically using the ``AS`` clause when
    generated, e.g. ``SELECT * FROM table AS aliasname``.

    Equivalent functionality is available via the
    :meth:`_expression.FromClause.alias`
    method available on all :class:`_expression.FromClause` objects.

    :param selectable: any :class:`_expression.FromClause` subclass,
        such as a table, select statement, etc.

    :param name: string name to be assigned as the alias.
        If ``None``, a name will be deterministically generated at compile
        time. Deterministic means the name is guaranteed to be unique against
        other constructs used in the same statement, and will also be the same
        name for each successive compilation of the same statement object.

    :param flat: Will be passed through to if the given selectable
     is an instance of :class:`_expression.Join` - see
     :meth:`_expression.Join.alias` for details.

    """
    return Alias._factory(selectable, name=name, flat=flat)


def cte(
    selectable: HasCTE, name: Optional[str] = None, recursive: bool = False
) -> CTE:
    r"""Return a new :class:`_expression.CTE`,
    or Common Table Expression instance.

    Please see :meth:`_expression.HasCTE.cte` for detail on CTE usage.

    """
    return coercions.expect(roles.HasCTERole, selectable).cte(
        name=name, recursive=recursive
    )


def except_(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return an ``EXCEPT`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    :param \*selects:
      a list of :class:`_expression.Select` instances.

    """
    return CompoundSelect._create_except(*selects)


def except_all(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return an ``EXCEPT ALL`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    :param \*selects:
      a list of :class:`_expression.Select` instances.

    """
    return CompoundSelect._create_except_all(*selects)


def exists(
    __argument: Optional[
        Union[_ColumnsClauseArgument[Any], SelectBase, ScalarSelect[Any]]
    ] = None,
) -> Exists:
    """Construct a new :class:`_expression.Exists` construct.

    The :func:`_sql.exists` can be invoked by itself to produce an
    :class:`_sql.Exists` construct, which will accept simple WHERE
    criteria::

        exists_criteria = exists().where(table1.c.col1 == table2.c.col2)

    However, for greater flexibility in constructing the SELECT, an
    existing :class:`_sql.Select` construct may be converted to an
    :class:`_sql.Exists`, most conveniently by making use of the
    :meth:`_sql.SelectBase.exists` method::

        exists_criteria = (
            select(table2.c.col2).
            where(table1.c.col1 == table2.c.col2).
            exists()
        )

    The EXISTS criteria is then used inside of an enclosing SELECT::

        stmt = select(table1.c.col1).where(exists_criteria)

    The above statement will then be of the form::

        SELECT col1 FROM table1 WHERE EXISTS
        (SELECT table2.col2 FROM table2 WHERE table2.col2 = table1.col1)

    .. seealso::

        :ref:`tutorial_exists` - in the :term:`2.0 style` tutorial.

        :meth:`_sql.SelectBase.exists` - method to transform a ``SELECT`` to an
        ``EXISTS`` clause.

    """  # noqa: E501

    return Exists(__argument)


def intersect(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return an ``INTERSECT`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    :param \*selects:
      a list of :class:`_expression.Select` instances.

    """
    return CompoundSelect._create_intersect(*selects)


def intersect_all(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return an ``INTERSECT ALL`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    :param \*selects:
      a list of :class:`_expression.Select` instances.


    """
    return CompoundSelect._create_intersect_all(*selects)


def join(
    left: _FromClauseArgument,
    right: _FromClauseArgument,
    onclause: Optional[_OnClauseArgument] = None,
    isouter: bool = False,
    full: bool = False,
) -> Join:
    """Produce a :class:`_expression.Join` object, given two
    :class:`_expression.FromClause`
    expressions.

    E.g.::

        j = join(user_table, address_table,
                 user_table.c.id == address_table.c.user_id)
        stmt = select(user_table).select_from(j)

    would emit SQL along the lines of::

        SELECT user.id, user.name FROM user
        JOIN address ON user.id = address.user_id

    Similar functionality is available given any
    :class:`_expression.FromClause` object (e.g. such as a
    :class:`_schema.Table`) using
    the :meth:`_expression.FromClause.join` method.

    :param left: The left side of the join.

    :param right: the right side of the join; this is any
     :class:`_expression.FromClause` object such as a
     :class:`_schema.Table` object, and
     may also be a selectable-compatible object such as an ORM-mapped
     class.

    :param onclause: a SQL expression representing the ON clause of the
     join.  If left at ``None``, :meth:`_expression.FromClause.join`
     will attempt to
     join the two tables based on a foreign key relationship.

    :param isouter: if True, render a LEFT OUTER JOIN, instead of JOIN.

    :param full: if True, render a FULL OUTER JOIN, instead of JOIN.

    .. seealso::

        :meth:`_expression.FromClause.join` - method form,
        based on a given left side.

        :class:`_expression.Join` - the type of object produced.

    """

    return Join(left, right, onclause, isouter, full)


def lateral(
    selectable: Union[SelectBase, _FromClauseArgument],
    name: Optional[str] = None,
) -> LateralFromClause:
    """Return a :class:`_expression.Lateral` object.

    :class:`_expression.Lateral` is an :class:`_expression.Alias`
    subclass that represents
    a subquery with the LATERAL keyword applied to it.

    The special behavior of a LATERAL subquery is that it appears in the
    FROM clause of an enclosing SELECT, but may correlate to other
    FROM clauses of that SELECT.   It is a special case of subquery
    only supported by a small number of backends, currently more recent
    PostgreSQL versions.

    .. seealso::

        :ref:`tutorial_lateral_correlation` -  overview of usage.

    """
    return Lateral._factory(selectable, name=name)


def outerjoin(
    left: _FromClauseArgument,
    right: _FromClauseArgument,
    onclause: Optional[_OnClauseArgument] = None,
    full: bool = False,
) -> Join:
    """Return an ``OUTER JOIN`` clause element.

    The returned object is an instance of :class:`_expression.Join`.

    Similar functionality is also available via the
    :meth:`_expression.FromClause.outerjoin` method on any
    :class:`_expression.FromClause`.

    :param left: The left side of the join.

    :param right: The right side of the join.

    :param onclause:  Optional criterion for the ``ON`` clause, is
      derived from foreign key relationships established between
      left and right otherwise.

    To chain joins together, use the :meth:`_expression.FromClause.join`
    or
    :meth:`_expression.FromClause.outerjoin` methods on the resulting
    :class:`_expression.Join` object.

    """
    return Join(left, right, onclause, isouter=True, full=full)


# START OVERLOADED FUNCTIONS select Select 1-10

# code within this block is **programmatically,
# statically generated** by tools/generate_tuple_map_overloads.py


@overload
def select(__ent0: _TCCA[_T0]) -> Select[Tuple[_T0]]:
    ...


@overload
def select(__ent0: _TCCA[_T0], __ent1: _TCCA[_T1]) -> Select[Tuple[_T0, _T1]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0], __ent1: _TCCA[_T1], __ent2: _TCCA[_T2]
) -> Select[Tuple[_T0, _T1, _T2]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
) -> Select[Tuple[_T0, _T1, _T2, _T3]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
    __ent5: _TCCA[_T5],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4, _T5]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
    __ent5: _TCCA[_T5],
    __ent6: _TCCA[_T6],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4, _T5, _T6]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
    __ent5: _TCCA[_T5],
    __ent6: _TCCA[_T6],
    __ent7: _TCCA[_T7],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4, _T5, _T6, _T7]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
    __ent5: _TCCA[_T5],
    __ent6: _TCCA[_T6],
    __ent7: _TCCA[_T7],
    __ent8: _TCCA[_T8],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8]]:
    ...


@overload
def select(
    __ent0: _TCCA[_T0],
    __ent1: _TCCA[_T1],
    __ent2: _TCCA[_T2],
    __ent3: _TCCA[_T3],
    __ent4: _TCCA[_T4],
    __ent5: _TCCA[_T5],
    __ent6: _TCCA[_T6],
    __ent7: _TCCA[_T7],
    __ent8: _TCCA[_T8],
    __ent9: _TCCA[_T9],
) -> Select[Tuple[_T0, _T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9]]:
    ...


# END OVERLOADED FUNCTIONS select


@overload
def select(*entities: _ColumnsClauseArgument[Any], **__kw: Any) -> Select[Any]:
    ...


def select(*entities: _ColumnsClauseArgument[Any], **__kw: Any) -> Select[Any]:
    r"""Construct a new :class:`_expression.Select`.


    .. versionadded:: 1.4 - The :func:`_sql.select` function now accepts
       column arguments positionally.   The top-level :func:`_sql.select`
       function will automatically use the 1.x or 2.x style API based on
       the incoming arguments; using :func:`_sql.select` from the
       ``sqlalchemy.future`` module will enforce that only the 2.x style
       constructor is used.

    Similar functionality is also available via the
    :meth:`_expression.FromClause.select` method on any
    :class:`_expression.FromClause`.

    .. seealso::

        :ref:`tutorial_selecting_data` - in the :ref:`unified_tutorial`

    :param \*entities:
      Entities to SELECT from.  For Core usage, this is typically a series
      of :class:`_expression.ColumnElement` and / or
      :class:`_expression.FromClause`
      objects which will form the columns clause of the resulting
      statement.   For those objects that are instances of
      :class:`_expression.FromClause` (typically :class:`_schema.Table`
      or :class:`_expression.Alias`
      objects), the :attr:`_expression.FromClause.c`
      collection is extracted
      to form a collection of :class:`_expression.ColumnElement` objects.

      This parameter will also accept :class:`_expression.TextClause`
      constructs as
      given, as well as ORM-mapped classes.

    """
    # the keyword args are a necessary element in order for the typing
    # to work out w/ the varargs vs. having named "keyword" arguments that
    # aren't always present.
    if __kw:
        raise _no_kw()
    return Select(*entities)


def table(name: str, *columns: ColumnClause[Any], **kw: Any) -> TableClause:
    """Produce a new :class:`_expression.TableClause`.

    The object returned is an instance of
    :class:`_expression.TableClause`, which
    represents the "syntactical" portion of the schema-level
    :class:`_schema.Table` object.
    It may be used to construct lightweight table constructs.

    :param name: Name of the table.

    :param columns: A collection of :func:`_expression.column` constructs.

    :param schema: The schema name for this table.

        .. versionadded:: 1.3.18 :func:`_expression.table` can now
           accept a ``schema`` argument.
    """

    return TableClause(name, *columns, **kw)


def tablesample(
    selectable: _FromClauseArgument,
    sampling: Union[float, Function[Any]],
    name: Optional[str] = None,
    seed: Optional[roles.ExpressionElementRole[Any]] = None,
) -> TableSample:
    """Return a :class:`_expression.TableSample` object.

    :class:`_expression.TableSample` is an :class:`_expression.Alias`
    subclass that represents
    a table with the TABLESAMPLE clause applied to it.
    :func:`_expression.tablesample`
    is also available from the :class:`_expression.FromClause`
    class via the
    :meth:`_expression.FromClause.tablesample` method.

    The TABLESAMPLE clause allows selecting a randomly selected approximate
    percentage of rows from a table. It supports multiple sampling methods,
    most commonly BERNOULLI and SYSTEM.

    e.g.::

        from sqlalchemy import func

        selectable = people.tablesample(
                    func.bernoulli(1),
                    name='alias',
                    seed=func.random())
        stmt = select(selectable.c.people_id)

    Assuming ``people`` with a column ``people_id``, the above
    statement would render as::

        SELECT alias.people_id FROM
        people AS alias TABLESAMPLE bernoulli(:bernoulli_1)
        REPEATABLE (random())

    :param sampling: a ``float`` percentage between 0 and 100 or
        :class:`_functions.Function`.

    :param name: optional alias name

    :param seed: any real-valued SQL expression.  When specified, the
     REPEATABLE sub-clause is also rendered.

    """
    return TableSample._factory(selectable, sampling, name=name, seed=seed)


def union(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return a ``UNION`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    A similar :func:`union()` method is available on all
    :class:`_expression.FromClause` subclasses.

    :param \*selects:
      a list of :class:`_expression.Select` instances.

    :param \**kwargs:
      available keyword arguments are the same as those of
      :func:`select`.

    """
    return CompoundSelect._create_union(*selects)


def union_all(
    *selects: _SelectStatementForCompoundArgument,
) -> CompoundSelect:
    r"""Return a ``UNION ALL`` of multiple selectables.

    The returned object is an instance of
    :class:`_expression.CompoundSelect`.

    A similar :func:`union_all()` method is available on all
    :class:`_expression.FromClause` subclasses.

    :param \*selects:
      a list of :class:`_expression.Select` instances.

    """
    return CompoundSelect._create_union_all(*selects)


def values(
    *columns: ColumnClause[Any],
    name: Optional[str] = None,
    literal_binds: bool = False,
) -> Values:
    r"""Construct a :class:`_expression.Values` construct.

    The column expressions and the actual data for
    :class:`_expression.Values` are given in two separate steps.  The
    constructor receives the column expressions typically as
    :func:`_expression.column` constructs,
    and the data is then passed via the
    :meth:`_expression.Values.data` method as a list,
    which can be called multiple
    times to add more data, e.g.::

        from sqlalchemy import column
        from sqlalchemy import values

        value_expr = values(
            column('id', Integer),
            column('name', String),
            name="my_values"
        ).data(
            [(1, 'name1'), (2, 'name2'), (3, 'name3')]
        )

    :param \*columns: column expressions, typically composed using
     :func:`_expression.column` objects.

    :param name: the name for this VALUES construct.  If omitted, the
     VALUES construct will be unnamed in a SQL expression.   Different
     backends may have different requirements here.

    :param literal_binds: Defaults to False.  Whether or not to render
     the data values inline in the SQL output, rather than using bound
     parameters.

    """
    return Values(*columns, literal_binds=literal_binds, name=name)



class ReportFileNotFound(Exception):

    def __init__(
        self,
        report_name: str,
    ) -> None:
        super(ReportFileNotFound, self).__init__(f"No JSON found for requested report: {report_name}")



import base64
from typing import Optional
from xml.etree import cElementTree as eTree

from helper import remove_xml_namespace
from messagebus import (
    MessageBus,
)
from order_api import OrderService


class OrderAPIRepository(object):

    def __init__(
        self,
        message_bus: MessageBus,
    ) -> None:

        self._message_bus = message_bus

    def get_order_service(
        self,
        pos_id: int,
    ) -> OrderService:

        return OrderService(
            message_bus=self._message_bus,
            pos_id=pos_id,
        )

    def get_order_picture(
        self,
        pos_id: int,
        order_id: int,
    ) -> Optional[eTree.Element]:

        order_service = self.get_order_service(pos_id=pos_id)
        return order_service.get_order_picture(order_id=order_id)

    def get_fiscal_xml(
        self,
        pos_id: int,
        order_id: int,
        order_xml: Optional[eTree.Element] = None,
    ) -> eTree.Element:

        if not order_xml:
            order_service = self.get_order_service(pos_id=pos_id)
            order_xml = order_service.get_order_picture(order_id=order_id)

        fiscal_xml_base_64 = order_xml.find(".//OrderProperty[@key='FISCAL_XML']").get("value")

        return remove_xml_namespace(xml=base64.b64decode(s=fiscal_xml_base_64).decode())



from typing import Dict

from application.model import UpdateType
from application.util import settings


class PosVersions(object):
    def __init__(
        self,
        last_versions,          # type: Dict[int, str]
        last_versions_dates,    # type: Dict[int, str]
        menu_version,           # type: str
        menu_name,              # type: str
    ):
        # type: (...) -> None

        catalog = last_versions[UpdateType.catalog.value] if UpdateType.catalog.value in last_versions else ""
        catalog_json_value = UpdateType.catalog_json.value
        catalog_json = last_versions[catalog_json_value] if catalog_json_value in last_versions else ""
        user = last_versions[UpdateType.user.value] if UpdateType.user.value in last_versions else ""
        loader = last_versions[UpdateType.loader.value] if UpdateType.loader.value in last_versions else ""
        binaries = last_versions[UpdateType.binaries.value] if UpdateType.binaries.value in last_versions else ""
        core = binaries.split("|")[0].split(":")[1] if binaries != "" else ""
        src = binaries.split("|")[1].split(":")[1] if binaries != "" else ""

        self.catalog = catalog
        self.catalog_json = catalog_json
        self.user = user
        self.src = src
        self.core = core
        self.loader = loader
        self.binaries = binaries if core != "" and src != "" else ""
        self.catalog_date = last_versions_dates[UpdateType.catalog.value] if catalog else ""
        self.user_date = last_versions_dates[UpdateType.user.value] if user else ""
        self.core_date = last_versions_dates[UpdateType.binaries.value] if core else ""
        self.src_date = last_versions_dates[UpdateType.binaries.value] if src else ""
        self.loader_date = last_versions_dates[UpdateType.loader.value] if loader else ""
        self.binaries_date = last_versions_dates[UpdateType.binaries.value] if core and src else ""
        self.catalog_json_date = last_versions_dates[catalog_json_value] if catalog_json else ""
        settings.check_status(self)
        self.update_control = settings.update_control
        self.menu_version = menu_version
        self.menu_name = menu_name



import sysactions

from externals.repository import AlertsRepository
from services.model import POSAlertNotificationRequest


class POSAlertNotificationInteractor(object):
    def __init__(
        self,
        alerts_repository,  # type: AlertsRepository
    ):
        # type: (...) -> None

        self.mb_context = sysactions.mbcontext
        self.alerts_repository = alerts_repository

    def process_request(
        self,
        request,  # type: POSAlertNotificationRequest
    ):
        # type: (...) -> None

        self.alerts_repository.set_sent_status(
            event_id=request.event_id,
            sent_status=1,
        )
